#include <AlignedJoy.h>
#include <Arduino.h>
#include "Arduino_LED_Matrix.h"
#include "NvmEeprom.hpp"

// joystick calibration
#define TIME_CAL_1 2000
#define TIME_CAL_2 3000

// static objects
AlignedJoy joystick_1(0, 1);  /// X on A0, Y on A1
ArduinoLEDMatrix ledMatrix;

/// Run calibration on the provided joystick. Save the results to the provided calibration
bool runJoystickCalibrationRoutine(AlignedJoy& joy, JoystickCalibration& cal) {
    /*
     * Start the joystick calibration in the center position. Use this method only if the
     * calibration of the axles is desired (axesCalibration).
     */
    // CALIBRATION 1
    Serial.print(
        "Center Joystick Calibration \n\nLeave the joystick in the centre position during the "
        "calibration\ntime which is set to ");
    Serial.print(TIME_CAL_1);
    Serial.println(" milliseconds.");
    delay(5000);
    Serial.println("Calibration started!");
    delay(500);
    /*
     * for calibrate the joystick center point use "centerCalibration" method;
     * objectname.centerCalibration(uint16_t timeOfCal). "timeOfCal" is the calibration time in
     * milliseconds
     */
    joy.middleCalibration(TIME_CAL_1);
    Serial.println("Joystick Centered!");
    delay(2000);

    // CALIBRATION 2
    /*
     * Calibration of the axes at the extreme points (min end max for each axis).
     * It is recommended to rotate the joystick in a circular way along its circumference throughout
     * the calibration time.
     */
    Serial.print(
        "Start axes Calibration \n\nRotate the joystick in a circular way along its "
        "circumference\nthroughout the calibration time (");
    Serial.print(TIME_CAL_2);
    Serial.println(" milliseconds).");
    delay(5000);
    Serial.println("Calibration started!");
    delay(500);
    /*
     * for calibrate the joystick axes points use "axesCalibration" method (bool type);
     * objectname.axesCalibration(uint16_t timeOfCal). "timeOfCal" is the calibration time in
     * milliseconds.
     */
    if (joy.axesCalibration(TIME_CAL_2)) {
        Serial.println("Calibration succesfull!");
    }
    else {
        Serial.println("Calibration failed!");
        return false;
    }

    // Print all points calibrated
    /*
     * You can use these values to save them to eeprom memory. In this way you will avoid requiring
     * the joystick calibration at each boot time. To set the parameters read by eeprom you have to
     * use the "setCalibratedPoint" method. If your project does not require the re-calibration of
     * the joystick then you can make a sketch like this only to display the calibrated values to
     * set them in the final project using the "setCalibratedPoint" method. To get the calibrated
     * point values use getCalibratedPoint(axis_t axis, point_t point). The parameters: the labels
     * of the "axis_t" shall be X and Y; the labels of "point_t" is MIN, MID and MAX.
     */
    Serial.print("X min -> ");
    Serial.print(joy.getCalibratedPoint(X, MIN));
    Serial.print(" | center ->  ");
    Serial.print(joy.getCalibratedPoint(X, MID));
    Serial.print(" | max -> ");
    Serial.println(joy.getCalibratedPoint(X, MAX));
    Serial.print("Y min -> ");
    Serial.print(joy.getCalibratedPoint(Y, MIN));
    Serial.print(" | center ->  ");
    Serial.print(joy.getCalibratedPoint(Y, MID));
    Serial.print(" | max -> ");
    Serial.println(joy.getCalibratedPoint(Y, MAX));

    // save values to cal struct
    cal.x.min = joy.getCalibratedPoint(X, MIN);
    cal.x.mid = joy.getCalibratedPoint(X, MID);
    cal.x.max = joy.getCalibratedPoint(X, MAX);
    cal.y.min = joy.getCalibratedPoint(Y, MIN);
    cal.y.mid = joy.getCalibratedPoint(Y, MID);
    cal.y.max = joy.getCalibratedPoint(Y, MAX);
    return true;
}

void setup() {
    // SERIAL INITIALIZE
    Serial.begin(115200);
    ledMatrix.loadSequence(LEDMATRIX_ANIMATION_STARTUP);
    ledMatrix.begin();
    while (!Serial) {
    }
    ledMatrix.play(true);

    // retrieve current calibration values
    NonVolatileMemory nvm = {};
    readNvm(nvm);

    // run joystick calibration if needed
    bool calibrationInvalid = nvm.valid();
    do {
        calibrationInvalid |=
            !nvm.calData.joystickCal.x.valid() || !nvm.calData.joystickCal.y.valid();
        if (calibrationInvalid) {
            calibrationInvalid |=
                !runJoystickCalibrationRoutine(joystick_1, nvm.calData.joystickCal);
        }
    } while (calibrationInvalid);
    // ensure calibration has been applied
    // TODO: API doesnt allow setting configured midpoint...
    joystick_1.setCalibratedPoint(axis_t::X, point_t::MIN, nvm.calData.joystickCal.x.min);
    joystick_1.setCalibratedPoint(axis_t::X, point_t::MID, nvm.calData.joystickCal.x.mid);
    joystick_1.setCalibratedPoint(axis_t::X, point_t::MAX, nvm.calData.joystickCal.x.max);
    joystick_1.setCalibratedPoint(axis_t::Y, point_t::MIN, nvm.calData.joystickCal.y.min);
    joystick_1.setCalibratedPoint(axis_t::Y, point_t::MID, nvm.calData.joystickCal.y.mid);
    joystick_1.setCalibratedPoint(axis_t::Y, point_t::MAX, nvm.calData.joystickCal.y.max);

    // run analog lever calibration if needed

    // write NVM in the event any values were updated
    writeNvm(nvm);
}

void loop() {
    // print joystick axes value
    Serial.print("joystick_1 X -> ");
    Serial.print(joystick_1.read(X));
    Serial.print(" | Y -> ");
    Serial.println(joystick_1.read(Y));
    delay(500);
}
